#include "config.h"
#include "main.h"
#include "pages.h"

#include <Updater.h>
#include "StoreCSV.h"
#include "RGBW.h"
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <WebSocketsServer.h>
#include <Preferences.h>

String convert_utf8_to_iso8859_1(String utf8) {
  String iso8859_1 = "";
  for (uint i = 0; i < utf8.length(); i++){
    char c = utf8[i]; // get the character
    if ((c & 0x80) == 0) { // is it a single-byte character?
      iso8859_1 += c; // if so, just append it
    } else if ((c & 0xE0) == 0xC0) { // is it a two-byte sequence?
      // extract the Unicode codepoint represented by the sequence
      int codepoint = ((c & 0x1F) << 6) | (utf8[++i] & 0x3F);
      if (codepoint <= 0xFF) { // is it within the ISO-8859-1 range?
        iso8859_1 += (char) codepoint; // if so, append the equivalent ISO-8859-1 character
      }
      // if not, we simply omit it
    } else {
      // it's a more-than-two-byte sequence, which is definitely outside the ISO-8859-1 range,
      // so we just skip the rest of the sequence
      if ((c & 0xF0) == 0xE0) i++;
      else if ((c & 0xF8) == 0xF0) i += 2;
      // if it's neither, it's a malformed sequence, so we skip it altogether
    }
  }
  return iso8859_1;
}

ESP8266WebServer server(80);
WebSocketsServer webSocket(81);

WiFiUDP UDP;
char UDP_packet[90 * 4] = {0};

WiFiUDP UDP2;
char UDP2_packet[60] = {0};

String ota_pw;

bool enableFallbackAP;

void handleWebClient() {
  server.handleClient();
}

void handleRoot() {
  server.send(200, "text/html", displayForm);
}

void tickerPage() {
  server.send(200, "text/html", tickerForm);
}

void pickerPage() {
  server.send(200, "text/html", colorPickerIndex);
}

void handleRainbow() {
  Serial.print("handleRainbow:");
  if (server.arg("rainbow") == "On") {
    rainbow = true;
    Serial.println("On");
    server.send(200, "text/html", "Turn Rainbow On");
  } else if (server.arg("rainbow") == "Off") {
    rainbow = false;
    Serial.println("Off");
    server.send(200, "text/html", "Turn Rainbow Off");
  }
  server.send(400, "text/html", "Error in handleRainbow, no value found " + server.arg("rainbow"));
}

void handleScroller() {
  Serial.print("handleScroller:");
  if (server.arg("scroller") == "On") {
    scroller = true;
    Serial.println("On");
    server.send(200, "text/html", "Turn Scroller On");
  } else if (server.arg("scroller") == "Off") {
    scroller = false;
    Serial.println("Off");
    server.send(200, "text/html", "Turn Scroller Off");
  }
  server.send(400, "text/html", "Error in handeScroller, no value found " + server.arg("scroller"));
}

char lookup_hex(char c) {
  switch (c & 0b00011111) {
    case '0':
      return 0x0;
    case '1':
      return 0x1;
    case '2':
      return 0x2;
    case '3':
      return 0x3;
    case '4':
      return 0x4;
    case '5':
      return 0x5;
    case '6':
      return 0x6;
    case '7':
      return 0x7;
    case '8':
      return 0x8;
    case '9':
      return 0x9;

    case 'a':
      return 0xA;
    case 'b':
      return 0xB;
    case 'c':
      return 0xC;
    case 'd':
      return 0xD;
    case 'e':
      return 0xE;
    case 'f':
      return 0xF;
  }

  return 0;
}

void handleSetPixels() {
  char hexbuf[720]; // NUM_PIX * bytes * 2

  Serial.print("handleSetPixels:");
  if (server.hasArg("pix")) {
    server.arg("pix").toCharArray(hexbuf, 720);

    uint8_t buf[360];
    uint8_t toggle = 0;
    for(int i = 0 ; i < 720 ; i++) {
      buf[i/2] = lookup_hex(hexbuf[i]) << (toggle * 4);

      toggle ^= 1;
    }

    Serial.println(hexbuf);
    server.send(200, "text/html", "Color");
  } else {
    server.send(400, "text/html", "Invalid request, no text found");
  }

}

void handleDisplayString() {
  String message;
  Serial.print("handleDisplayString:");
  if (server.hasArg("text")) {
    message = server.arg("text");
 
    writeText(message.c_str());
 
    Serial.println(message);
    server.send(200, "text/html", "Display: '" + message +"'");
  } else {
    server.send(400, "text/html", "Invalid request, no text found");
  }
}

std::vector<String> ticker_lines;
uint32_t ticker_timeout;

/*
void handleTickerEvent() {
    time_t now;

    if (ticker_lines.empty()) {
      return;
    }

    static unsigned long lastLoopTime = millis();
    static int idx = 0;
    if(millis() - lastLoopTime < ticker_timeout) {
        return;
    }
    lastLoopTime = millis();

    if (idx < ticker_lines.size())
    {
      SoftSerial.write(0x0A);
      SoftSerial.write(0x09);
      SoftSerial.write(0x0D);
      SoftSerial.print(ticker_lines[idx]);
      idx++;

    } else {
      idx = 0;
    }
}
*/

void handleOTA() {
  server.sendHeader("Connection", "close");
  server.send(200, "text/html", loginIndex);
}

void handleServerIndex() {
  server.sendHeader("Connection", "close");
  if(server.arg("k") != ota_pw)
    server.send(400, "text/html", "invalid credentials.");
  else
    server.send(200, "text/html", serverIndex);
}

void handleUpdate() {
  server.sendHeader("Connection", "close");
  if(server.arg("k") != ota_pw) {
    server.send(400, "text/html", "invalid credentials.");
  } else {
    server.send(200, "text/html", serverIndex);
    server.send(200, "text/plain", (Update.hasError()) ? "FAIL" : "OK");
    ESP.restart();
  }
}

void handleUpload() {
  HTTPUpload& upload = server.upload();
  size_t bytediv = 127 / max(upload.contentLength, size_t(1)); // bytes / step; upload / 127
  fillColor(0xFF, 0xA5, 0);
  writeText("upgrading firmware...");

  if (upload.status == UPLOAD_FILE_START) {
    Serial.printf("Update: %s\n", upload.filename.c_str());
    if (!Update.begin(upload.contentLength)) {
      Update.printError(Serial);
    }
  } else if (upload.status == UPLOAD_FILE_WRITE) {
    if (Update.write(upload.buf, upload.currentSize) != upload.currentSize) {
      Update.printError(Serial);
    }
  } else if (upload.status == UPLOAD_FILE_END) {
    if (Update.end(true)) {
      Serial.printf("Update Success: %u\nRebooting...\n", upload.totalSize);

      fillColor(0, 255, 0);
      writeText("upgraded firmware!");
    } else {
      Update.printError(Serial);
    }
  }
}

void handleCredentials() {
    String ssid;
    String wifi_pw;
    if (server.hasArg("ssid") && server.hasArg("wifi_pw")) {
      ssid = server.arg("ssid");
      wifi_pw = server.arg("wifi_pw");
      ota_pw = server.arg("ota_pw");

      prefs.begin("wifi", false);
      prefs.putString("ssid", ssid);
      prefs.putString("wifi_pw", wifi_pw);
      prefs.putString("ota_pw", ota_pw);
      prefs.end();

      server.send(200, "text/html", "Saved credentials!");
    } else {
      server.send(400, "text/plain", "Bad Request");
    }
}

void CSVUploadPage() {
  server.sendHeader("Connection", "close");
  server.send(200, "text/html", csvUploadPage);
}

void webSocketEvent(uint8_t num, WStype_t type, uint8_t * payload, size_t length){
  if(type == WStype_TEXT){
    String message = String((char*) payload);
    Serial.print("WS:");
    Serial.println(message);
    
    writeText(message.c_str());
  } else if(type == WStype_BIN) {
    for(size_t i = 0 ; i < length ; i+=4 ) {
      set_pixelw(i/4, CRGBW(payload[i], payload[i+1], payload[i+2], payload[i+3]));
    }

    show_pixel();
  }
}

void handleFallbackRoot() {
  Serial.print("handleFallbackRoot");
  prefs.begin("wifi", true);
  String ssid = prefs.getString("ssid", "");
  String wifi_pw = prefs.getString("wifi_pw", "");
  String ota_pw = prefs.getString("ota_pw", "");
  prefs.end();

  char wifiCredentialsPage[1024];
  snprintf(wifiCredentialsPage, 1024, wifiCredentialsTemplate, ssid.c_str(), wifi_pw.c_str(), ota_pw.c_str());

  server.send(200, "text/html", wifiCredentialsPage);
}

void startAPMode() {
  progress(4, SETUP_PROGRESS_MAX, 255, 255, 0);

  WiFi.disconnect();
  WiFi.softAPConfig(FALLBACK_IP, GATEWAY_IP, SUBNET_IP);

  add_progress();
  WiFi.mode(WIFI_AP);
  WiFi.softAP(FALLBACK_SSID, FALLBACK_PASSWORD); // Using default credentials

  IPAddress myIP = WiFi.softAPIP();
  Serial.print("AP IP address: ");
  Serial.println(myIP);

  server.on("/", handleFallbackRoot);
  server.on("/saveCredentials", HTTP_POST, handleCredentials);

  enableFallbackAP = true;

  server.begin();
  add_progress();
}

void connectToWiFi() {
  add_progress();
  writeText("Connecting To Wifi");
  esp_delay(1000);

  enableFallbackAP = false;

  // Attempt to connect to Wi-Fi with saved credentials
  prefs.begin("wifi", true);  // Begin preferences in read-only mode
  String ssid = prefs.getString("ssid", "");
  const char *cstrssid = ssid.c_str();
  String wifi_pw = prefs.getString("wifi_pw", "");
  ota_pw = prefs.getString("ota_pw", "");
  prefs.end();

  if (ssid != "" && wifi_pw != "") {
    writeText(cstrssid);

    WiFi.begin(cstrssid, wifi_pw.c_str());

    for(int i = 0 ; i < 5 ; i++ ) {
      if(WiFi.isConnected()) {
        break;
      }

      esp_delay(500);

      add_progress();
      show_progress(); // hopefully fixes some shit
    }
    
    if (WiFi.waitForConnectResult() != WL_CONNECTED) {
      Serial.println("Failed to connect to Wi-Fi with saved credentials. Starting AP mode.");
      enableFallbackAP = true;
      
      writeText("invalid credentials");
      esp_delay(5000);

      startAPMode();
    }
  } else {
    Serial.println("No saved credentials found. Starting AP mode.");
    enableFallbackAP = true;

    fillColor(255, 0, 0);
    writeText("no credentails");
    esp_delay(5000);

    startAPMode();
  }

  esp_delay(1000);

  if(enableFallbackAP) {
    esp_delay(1000);
  }

  set_progress(6, SETUP_PROGRESS_MAX);

  IPAddress myIP = enableFallbackAP ? FALLBACK_IP : WiFi.localIP();
  String ip_adress_s = myIP.toString();

  Serial.print("\nConnected to ");
  Serial.println(ssid);
  Serial.print("IP address: ");
  Serial.println(ip_adress_s.c_str());

  /*use mdns for host name resolution*/
  if (!MDNS.begin(HOST))
    Serial.println("Error setting up MDNS responder, MDNS will not be used");
  else
    Serial.println("mDNS responder started");

  add_progress();

  #define CONNECTED "Connected:"
  #define FALLBACK "Fallback: "

  if(enableFallbackAP) {
    writeText(FALLBACK FALLBACK_SSID);
  } else {
    char buf[strlen(CONNECTED) + strlen(cstrssid)] = CONNECTED;
    strcat(buf, cstrssid);

    writeText(buf);
  }
  esp_delay(2000);
  const char* cstrip = ip_adress_s.c_str();

  if(enableFallbackAP) {
    char buf[strlen(FALLBACK) + strlen(cstrip)] = FALLBACK;
    strcat(buf, cstrip);

    writeText(buf);
  } else {
    char buf[strlen(CONNECTED) + strlen(cstrip)] = CONNECTED;
    strcat(buf, cstrip);

    writeText(buf);
  }
  add_progress();

  esp_delay(2000);
  
  if(enableFallbackAP) {
    writeText(FALLBACK FALLBACK_SSID);
  }
}

void setupWebServer() {
  server.on("/", handleRoot);
  server.on("/ticker", HTTP_GET, tickerPage);
  server.on("/pickerForm", HTTP_GET, pickerPage);
//  server.on("/tickerForm", HTTP_POST, handleTicker);
  server.on("/Scroller", HTTP_POST, handleScroller);
  server.on("/Rainbow", HTTP_POST, handleRainbow);
  server.on("/StringForm", HTTP_POST, handleDisplayString);
  server.on("/SetPixels", HTTP_POST, handleSetPixels);
  server.on("/OTA", HTTP_GET, handleOTA);
  server.on("/serverIndex", HTTP_GET, handleServerIndex);
  server.on("/uploadCSV", HTTP_GET, CSVUploadPage);
  server.on("/uploadCSV", HTTP_POST, []() {
    server.send(200, "text/plain", "CSV Uploaded Successfully");
  }, handleCSVUpload);

  server.on("/update", HTTP_POST, []() {
    handleUpdate();
  }, handleUpload);
  server.begin();

  UDP.begin(UDP_PORT);
  UDP2.begin(UDP_PORT + 1);
  Serial.print("Listening on UDP port ");
  Serial.println(UDP_PORT);

  webSocket.onEvent(webSocketEvent);
  webSocket.begin();

}

void handleUDP() {
  int packetSize = UDP.parsePacket();
  if (packetSize) {
    Serial.print("Received packet! Size: ");
    Serial.println(packetSize); 
    int len = UDP.read(UDP_packet, 90 * 4);

    for(int i = 0 ; i < len ; i += 4 )
      set_pixelw(i/4, CRGBW(UDP_packet[i],UDP_packet[i+1],UDP_packet[i+2],UDP_packet[i+3]));

    show_pixel();
  }

  packetSize = UDP2.parsePacket();
  if (packetSize) {
    Serial.print("Received packet2! ize: ");
    Serial.println(packetSize); 
    int len = UDP2.read(UDP2_packet, 60-1);

    UDP2_packet[len] = '\0';

    writeText(UDP2_packet);
  }
}