#include "main.h"

#include <Arduino.h>
#include <FastLED.h>
#include <RGBW.h>
#include <SoftwareSerial.h>

#include <config.h>
#include <Server.h>

#include <Preferences.h>
Preferences prefs = Preferences();

SoftwareSerial fis(D2, D1, false);
#include <FIS.c>

// toggle for the build-in text scroller
bool scroller = true;
bool rainbow = true;

CRGBW leds[NUM_LEDS];
CRGB *ledsRGB = (CRGB *) &leds[0];

void sleep(uint mils) {
  Serial.print("sleep");
  Serial.println(mils);

  uint end = millis() + mils;
  
  while(end > millis()) {
    delay(10);
  }
}

void setup() {
  Serial.begin(115200);

  Serial.println("FastLED");
  FastLED.addLeds<WS2812B, DATA_PIN, RGB>(ledsRGB, getRGBWsize(NUM_LEDS));
  FastLED.setBrightness(255);

  // databits 5 + (x & 07) -> x = 2
  // docu would be nice
  // 7 bits
  // even parity
  // 2 stopbits?
  fis.begin(1200, SWSERIAL_7E2);

  fillColorW(255, 255, 255, 127);
  FastLED.show();

  sleep(100);

  writeText("booting, hi :3");
  Serial.println("boot!");

  sleep(200);

  progress(0, SETUP_PROGRESS_MAX, 0, 254, 0);

  // wifi stuff
  connectToWiFi();

  setupWebServer();
}

time_t next = 0;
time_t now;
uint counter = 0;
uint8_t hue;

int analogVal = 0;

void loop() {
  handleUDP();

  handleWebClient();
  if (enableFallbackAP) {
    return;
  }

  webSocket.loop();

  now = millis();

  if ( now > next) {
    next = now + 5;
  } else {
    return;
  }

  hue = uint8_t(counter/2);

  // rainbow right to left
  if(rainbow) {
    for(int i = 0 ; i < NUM_LEDS/2 ; i++ ) {
      uint8_t chur = hue + ( HSVM * i );

      leds[i] = CHSV(chur, 255, 255);
      leds[NUM_LEDS-i] = CHSV(chur, 255, 255);
    } //*/

    FastLED.show();
  }
  // rainbow center out btoked
  /*
  int li;
  CRGB c;
  for(int i = 0 ; i < NUM_LEDS/4 ; i++ ) {
    uint8_t chur = hue + ( HSVM * i );
    li = NUM_LEDS/4 - li;
    c = CHSV(chur, 255, 255);

    leds[li] = c;
    leds[NUM_LEDS/2 - li] = c;

    leds[NUM_LEDS/2 + li] = c;
    leds[NUM_LEDS-li] = c;
  }
*/
  // loop di loop
  /*
  for(int i = 0 ; i < NUM_LEDS ; i++ )
    leds[i] = CRGBW(0,0,0,255);

  for(int i = counter%NUM_LEDS ; i < counter%NUM_LEDS + 7 ; i++ ) {
    leds[i%NUM_LEDS] = CHSV(hue, 255,255);
  }
  */

  /*
  fill_stripes(CRGB(91, 206, 250), CRGB(245, 169, 184), CRGB(255, 255, 255),
    CRGB(245, 169, 184), CRGB(91, 206, 250));
  */

  // 1000ms / 5ms -> 10s
  if (scroller && ((counter % (10000/5)) == 0)) {
    char *a = lines[random(0, num_lines)];

    Serial.print("@");
    Serial.print(counter);
    Serial.print(">");
    Serial.println(a);

    writeText(a);
  }

  counter ++;
}

void writeText(const char text[]) {
    // Example string to be sent
    // header + terminator + checksum
    int strl = strlen(text);

    int len_t = 1 + 1 + 1;
    len_t += 16; // at least for padding thingy

    if ( strl > 16 )
      len_t += strl - 16;

    // Create a buffer with a size of 16 bytes
    char b[len_t];

    // Append 'v' to the beginning
    b[0] = 'v';

    // Append the string t to the buffer
    memcpy(b+1, text, len_t - 1);

    // Fill the remaining buffer with spaces (0x20)
    // from after start til after string before checksum
    for (int i = 1 + strl; i < len_t-2; i++) {
        b[i] = 0x20;
    }

    b[len_t-2] = 0x0D; // trailer

    char checksum = 0x7F;
    
    // except checksum itself lel
    for ( int i = 0 ; i < len_t-1 ; i++ )
      checksum ^= b[i];

    b[len_t-1] = checksum; // checkum

    // Send the buffer over SoftwareSerial
    fis.write(b, len_t);
    //for ( uint8 i = 0 ; i < len_t ; i++ )
    //  Serial.printf("d %d %x %c\n", i, b[i], b[i]);
}

/*
 =======
    -------              Fast LED stuff
 =======
*/

void fill_stripes(CRGB a, CRGB b, CRGB c, CRGB d, CRGB e) {
  #define stripewidth (NUM_LEDS / 5 / 2)

  #define STRIPE(I, VARIABLE) \
    fillColorFT(stripewidth * I,            stripewidth * (I+1),            VARIABLE); \
    fillColorFT(NUM_LEDS - (stripewidth * (I+1)), NUM_LEDS - (stripewidth * I), VARIABLE);

  STRIPE(0, a)
  STRIPE(1, b)
  STRIPE(2, c)
  STRIPE(3, d)
  STRIPE(4, e)
}

void set_pixelw(int i, CRGBW color) {
  leds[i] = color;
}

void show_pixel() {
  FastLED.show();
}

void fillColorFT(int from, int to, CRGB color) {
  for ( int i = from ; i < to ; i++ )
    leds[i] = color;
}

void fillColorW(uint8_t r, uint8_t g, uint8_t b, uint8_t w) {
    for ( int i = 0 ; i < NUM_LEDS ; i++ )
        leds[i] = CRGBW(r, g, b, w);

    FastLED.show();
}

void fillColor(uint8_t r, uint8_t g, uint8_t b) {
    for(int i = 0 ; i < NUM_LEDS ; i++ )
        leds[i] = CRGB(r,g,b);

    FastLED.show();
}

uint8_t progress_status = 0;
uint8_t progress_total = 1;
uint8_t progress_r = 0;
uint8_t progress_g = 0;
uint8_t progress_b = 0;

#define PROGRESS_BACKGROUND CRGBW(0,0,0,254);

// sets progress thingy
void progress_color(uint8 r, uint8 g, uint8 b) {
  progress_r = r; progress_g = g; progress_b = b;
}

void add_progress() {
  progress(progress_status+1, progress_total,
    progress_r, progress_g, progress_b);
}

void set_progress(uint8 x, uint8 of) {
  progress(x, of, progress_r, progress_g, progress_b);
}

void show_progress() {
  progress(progress_status, progress_total,
    progress_r, progress_g, progress_b);
}

void progress(uint8 progress, uint8 total, uint8 r, uint8 g, uint8 b) {
  Serial.print(progress);
  Serial.print("/");
  Serial.println(total);

  progress_status = progress;
  progress_total = total;

  progress_r = r; progress_g = g; progress_b = b;

  CRGB c = CRGB(r, g, b);

  total = max(total, uint8(1));

  uint8_t lastled = progress * NUM_LEDS/2 / total;

    for(int i = 0 ; i < NUM_LEDS ; i++ )
        leds[i] = PROGRESS_BACKGROUND;

    for(int i = 0 ; i < lastled ; i++ ) {
        leds[i] = c;
        leds[NUM_LEDS - i] = c;
    }

    FastLED.show();
}